import json
import os

def load_config(path):
    if os.path.exists(path):
        with open(path, 'r') as f:
            return json.load(f)
    else:
        print(f"Config file not found. ({path})")


import argparse
import colorama
import requests
import justlog
import json
import os
from raincoat import shared
from .helpers import greet, get_torrent_by_id, fetch_torrent_url
from tabulate import tabulate
from .torrent import torrent, filter_out, transmission, deluge, qbittorrent, local
from justlog import justlog, settings
from justlog.classes import Severity, Output, Format
from .config import load_config
from pathlib import Path

parser = argparse.ArgumentParser()
parser.add_argument("search", help="What to search for.")
parser.add_argument("-k", "--key", help="The Jackett API key.")
parser.add_argument("-l", "--length", help="Max results description length.", type=int)
parser.add_argument("-L", "--limit", help="Max number of results.", type=int)
parser.add_argument("-c", "--config", help="Specify a different config file path.")
parser.add_argument("-s", "--sort", help="Change sorting criteria.", action="store", dest="sort", choices=['seeders', 'leechers', 'ratio', 'size', 'description'])
parser.add_argument("-i", "--indexer", help="The Jackett indexer to use for your search.")
parser.add_argument("--local", help="Override torrent provider with local download.", action="store_true")
parser.add_argument("--verbose", help="Very verbose output to logs.", action="store_true")
args = parser.parse_args()

shared.init()

# Use default path for the config file and load it initially
cfg_path = f"{str(Path.home())}/.config/{shared.APP_NAME}.json"
if args.config is not None:
    cfg_path = args.config
cfg = load_config(cfg_path)

shared.TORRENTS = []
shared.APIKEY = cfg['jackett_apikey']
shared.JACKETT_URL = cfg['jackett_url']
shared.JACKETT_INDEXER = cfg['jackett_indexer']
shared.DESC_LENGTH = cfg['description_length']
shared.EXCLUDE = cfg['exclude']
shared.RESULTS_LIMIT = cfg['results_limit']
shared.CLIENT_URL = cfg['client_url']
shared.DISPLAY = cfg['display']
shared.TOR_CLIENT = cfg['torrent_client']
shared.TOR_CLIENT_USER = cfg['torrent_client_username']
shared.TOR_CLIENT_PW = cfg['torrent_client_password']
shared.DOWNLOAD_DIR = cfg['download_dir']
shared.CURRENT_PAGE = 0
shared.VERBOSE_MODE = False


# Setup logger
logger = justlog.Logger(settings.Settings())
logger.settings.colorized_logs = True
logger.settings.log_output = [Output.FILE]
logger.settings.log_format = Format.TEXT
logger.settings.log_file = f"{str(Path.home())}/.config/{shared.APP_NAME}.log"
logger.settings.update_field("timestamp", "$TIMESTAMP")
logger.settings.update_field("level", "$CURRENT_LOG_LEVEL")
logger.settings.string_format = "[ $timestamp ] :: $CURRENT_LOG_LEVEL :: $message"

logger.debug(f"{shared.APP_NAME} v{shared.VERSION}")
greet(shared.VERSION)

def set_overrides():
    if args.key is not None:
        shared.APIKEY = args.key       

    if args.length is not None:        
        shared.DESC_LENGTH = args.length

    if args.limit is not None:        
        shared.RESULTS_LIMIT = args.limit

    if args.indexer is not None:        
        shared.JACKETT_INDEXER = args.key

    # Set default sorting
    if args.sort is None:
        args.sort = "seeders"

    if args.local:
        shared.TOR_CLIENT = "local"
    
    if args.verbose:
        shared.VERBOSE_MODE = True

def prompt_torrent():
    print("\nCommands: \n\t:download, :d ID\n\t:next, :n\n\t:prev, :p\n\t:quit, :q\n\tTo search something else, just type it and press enter.")
    try:
        cmd = input("-> ")
    except Exception as e:
        print(f"Invalid input: {str(e)}")
        prompt_torrent()
    if cmd.startswith(":download") or cmd.startswith(":d"):
        if len(cmd.split()) < 2:
            print("Invalid input")
            prompt_torrent()
        id = cmd.split()[1]
        if not id.isdigit():
            print(f"Not a valid id.({id})")
            logger.warning(f"Not a valid id.({id}). We were expecting an integer.")
            prompt_torrent()
        else:
            download(id)
            exit()
    if cmd.startswith(":quit") or cmd.startswith(":q"):
        exit()
    if cmd.startswith(":next") or cmd.startswith(":n"):
        display_results(shared.CURRENT_PAGE + 1)
    if cmd.startswith(":prev") or cmd.startswith(":p"):
        display_results(shared.CURRENT_PAGE - 1)        
    if cmd.strip() == "":
        prompt_torrent()
    search(cmd)

def download(id):
    torrent = get_torrent_by_id(shared.TORRENTS, id)
    if torrent is None:
        print(f"Cannot find {id}.")
        logger.warning(f"Invalid id. The ID provided was not found in the list.")
        search(args.search)    
    else:    
        if shared.TOR_CLIENT.lower() == "transmission":
            transmission(torrent, shared.CLIENT_URL, shared.TOR_CLIENT_USER, shared.TOR_CLIENT_PW, logger)
        elif shared.TOR_CLIENT.lower() == "deluge":
            deluge(torrent, shared.CLIENT_URL, shared.TOR_CLIENT_USER, shared.TOR_CLIENT_PW, logger)
        elif shared.TOR_CLIENT.lower() == "qbittorrent":
            qbittorrent(torrent, shared.CLIENT_URL, shared.TOR_CLIENT_USER, shared.TOR_CLIENT_PW, logger)
        elif shared.TOR_CLIENT.lower() == "local":
            local(torrent, shared.DOWNLOAD_DIR, logger)            
        else:
            print(f"Unsupported torrent client. ({shared.TOR_CLIENT})")
            exit()

def search(search_terms):
    print(f"Searching for \"{search_terms}\"...\n")
    try:
        url = f"{shared.JACKETT_URL}/api/v2.0/indexers/{shared.JACKETT_INDEXER}/results?apikey={shared.APIKEY}&Query={search_terms}"
        r = requests.get(url)
        logger.debug(f"Request made to: {url}")
        logger.debug(f"{str(r.status_code)}: {r.reason}")
        logger.debug(f"Headers: {json.dumps(dict(r.request.headers))}")
        if r.status_code != 200:
            print(f"The request to Jackett failed. ({r.status_code})")
            logger.error(f"The request to Jackett failed. ({r.status_code}) :: {shared.JACKETT_URL}api?passkey={shared.APIKEY}&search={search_terms}")
            exit()
        res = json.loads(r.content)
        res_count = len(res['Results'])
        logger.debug(f"Search yielded {str(res_count)} results.")
        if shared.VERBOSE_MODE:
            logger.debug(f"Content: {r.content.decode()}")
    except Exception as e:
        print(f"The request to Jackett failed.")
        logger.error(f"The request to Jackett failed. {str(e)}")
        exit()
    id = 1

    for r in res['Results']:
        if filter_out(r['Title'], shared.EXCLUDE):
            continue
        if len(r['Title']) > shared.DESC_LENGTH:
            r['Title'] = r['Title'][0:shared.DESC_LENGTH]
        download_url = r['MagnetUri'] if r['MagnetUri'] else r['Link']
        shared.TORRENTS.append(torrent(id, r['Title'].encode(
            'ascii', errors='ignore'), r['CategoryDesc'], r['Seeders'], r['Peers'], download_url, r['Size']))
        id += 1    

    # Sort torrents array
    sort_torrents(shared.TORRENTS)

    # Display results
    shared.CURRENT_PAGE = 1
    display_results(shared.CURRENT_PAGE)

def display_results(page):
    display_table = []
    if page < 1:
        prompt_torrent()    
    shared.CURRENT_PAGE = page
    count = 0
    slice_index = (shared.CURRENT_PAGE - 1) * shared.RESULTS_LIMIT
    for tor in shared.TORRENTS[slice_index:]:
        if count >= shared.RESULTS_LIMIT:
            break
        tor.size = "{:.2f}".format(float(tor.size)/1000000)
        display_table.append([tor.id, tor.description, tor.media_type,
                              f"{tor.size}GB", tor.seeders, tor.leechers, tor.ratio])
        count += 1
    print(tabulate(display_table, headers=[    
          "ID", "Description", "Type", "Size", "Seeders", "Leechers", "Ratio"], floatfmt=".2f", tablefmt=shared.DISPLAY))
    print(f"\nShowing page {shared.CURRENT_PAGE} - ({count * shared.CURRENT_PAGE} of {len(shared.TORRENTS)} results), limit is set to {shared.RESULTS_LIMIT}")    
    prompt_torrent()

def sort_torrents(torrents):
    if args.sort == "seeders":
        return torrents.sort(key=lambda x: x.seeders, reverse=True)
    if args.sort == "leechers":
        return torrents.sort(key=lambda x: x.leechers, reverse=True)        
    if args.sort == "size":
        return torrents.sort(key=lambda x: x.size, reverse=True)
    if args.sort == "ratio":
        return torrents.sort(key=lambda x: x.ratio, reverse=True)
    if args.sort == "description":
        return torrents.sort(key=lambda x: x.description, reverse=True)

def main():
    set_overrides()
    search(args.search)



# Globals
def init():
    global VERSION
    VERSION = "0.8"
    global APP_NAME
    APP_NAME = "Raincoat"
    global TORRENTS
    TORRENTS = []
    global APIKEY
    APIKEY = ""
    global JACKETT_URL
    JACKETT_URL = ""
    global JACKETT_INDEXER
    JACKETT_INDEXER = ""
    global DESC_LENGTH
    DESC_LENGTH = ""
    global EXCLUDE
    EXCLUDE = ""
    global RESULTS_LIMIT
    RESULTS_LIMIT = ""
    global CLIENT_URL
    CLIENT_URL = ""
    global DISPLAY
    DISPLAY = ""
    global TOR_CLIENT
    TOR_CLIENT = ""
    global TOR_CLIENT_USER
    TOR_CLIENT_USER = ""
    global TOR_CLIENT_PW
    TOR_CLIENT_PW = ""
    global DOWNLOAD_DIR
    DOWNLOAD_DIR = ""
    global  CURRENT_PAGE
    CURRENT_PAGE = 0
    global VERBOSE_MODE
    VERBOSE_MODE = False







import requests
import os.path as pt
import shutil
import tempfile
import sys
import json
from raincoat import shared as shared
from justlog import justlog, settings
from justlog.classes import Severity, Output, Format
from pathlib import Path
from time import sleep

# Setup logger
logger = justlog.Logger(settings.Settings())
logger.settings.colorized_logs = True
logger.settings.log_output = [Output.FILE]
logger.settings.log_format = Format.TEXT
logger.settings.log_file = f"{str(Path.home())}/.config/Raincoat.log"
logger.settings.update_field("timestamp", "$TIMESTAMP")
logger.settings.update_field("level", "$CURRENT_LOG_LEVEL")
logger.settings.string_format = "[ $timestamp ] :: $CURRENT_LOG_LEVEL :: $message"

def greet(VERSION):
    print(r"")
    print(r", // ,,/ ,.// ,/ ,// / /, // ,/, /, // ,/,, // ,,/ ,./")
    print(r"_/_/,/,_,_/, /, /,.__,/,,/ ,.// ,/ ,// / /, // ,__ , /")
    print(r"\______   \_,_/_  |__| ,_,_,/,_/,_  _,,/, //_,_/  |_/,")
    print(r",|       _/\__  \ |  |/    \_/ ___\/  _ \__  \\   __\ ")
    print(r"/|    |   \ / __ \|  |   |  \  \__(  <_> ) __ \|  |,,/")
    print(r",|____|_  /(____  /__|___|  /\___  >____(____  /__|, ,")
    print(rf",/ ,,/,/\/,,/, ,\/ ,// , /\/,/ ,,\/, ,, // /,\/ {VERSION}")
    print(r", // ,/,// ,/ ,, ,/, // ,/ /,// ,/ ,// ,/ ,, /,//, /,/")
    print("")


def get_torrent_by_id(torrents, tid):
    for torrent in torrents:
        if torrent.id == int(tid):
            return torrent
    return None

def fetch_torrent_url(torrent):
    try:
        r = requests.get(torrent.download, allow_redirects=False)
        logger.debug(f"Requesting {torrent.download}")
        logger.debug(f"{str(r.status_code)}: {r.reason}")
        logger.debug(f"Headers: {json.dumps(dict(r.request.headers))}")
        if shared.VERBOSE_MODE:
            logger.debug(f"Content: {r.content.decode()}")

        if r.status_code == 302:
            if r.headers['Location'] is not None:
                return r.headers['Location']
            else:
                logger.error(f"Bad headers in torrent: ({r.headers})")
        else:
            logger.error(f"Unexpected return code: {r.status_code}")
    except Exception as e:
        logger.error(f"Could not fetch torrent url: {str(e)}")
        exit()

def convert_to_torrent(url, save_path):
    # Importing here to prevent unneeded dependecies
    import libtorrent as lt

    if not pt.isdir(save_path):
        print(f"Invalid output folder: {save_path}")
        sys.exit(0)

    tempdir = tempfile.mkdtemp()
    ses = lt.session()
    params = {
        'save_path': tempdir,
        'storage_mode': lt.storage_mode_t(2),
        'url': url
    }
    handle = ses.add_torrent(params)

    print("Downloading Metadata (this may take a while)")
    while (not handle.has_metadata()):
        try:
            sleep(1)
        except KeyboardInterrupt:
            print("Aborting...")
            ses.pause()
            print("Cleanup dir " + tempdir)
            shutil.rmtree(tempdir)
            sys.exit(0)
    ses.pause()

    torinfo = handle.get_torrent_info()
    torfile = lt.create_torrent(torinfo)

    output = pt.abspath(torinfo.name() + ".torrent")
    output = pt.abspath(pt.join(save_path, torinfo.name() + ".torrent"))

    print(F"Saving torrent file to: {output}...")
    f = open(output, "wb")
    f.write(lt.bencode(torfile.generate()))
    f.close()
    print("Cleaning up temp files...")
    ses.remove_torrent(handle)
    shutil.rmtree(tempdir)



import requests
from clutch import Client as tClient
from deluge_client import DelugeRPCClient
from qbittorrent import Client as qClient
from .helpers import fetch_torrent_url, convert_to_torrent

class torrent:
    def __init__(self, id, description, media_type, seeders, leechers, download, size):
        self.id = id
        self.media_type = media_type
        self.description = description
        self.seeders = seeders
        self.leechers = leechers
        self.download = download
        if(leechers > 0):
            self.ratio = self.seeders / self.leechers
        else:
            self.ratio = self.seeders
        self.size = size / 1000

def filter_out(title, exclusions):
    exclusions = exclusions.split()
    for exclude in exclusions:
        if exclude in title.lower():
            return True
    return False

def deluge(torrent, CLIENT_URL, TOR_CLIENT_USER, TOR_CLIENT_PW, logger):
    TOR_CLIENT = "Deluge"
    print(f"Sending {torrent.description.decode('ascii')} to {TOR_CLIENT}")
    url = fetch_torrent_url(torrent)
    try:
        logger.debug("Connecting to torrent client...")
        # Connection
        logger.debug(f"{TOR_CLIENT} connection info: {CLIENT_URL.split(':')[0]}, {int(CLIENT_URL.split(':')[1])}, {TOR_CLIENT_USER}")
        client = DelugeRPCClient(CLIENT_URL.split(':')[0], int(CLIENT_URL.split(':')[1]), TOR_CLIENT_USER, TOR_CLIENT_PW)
        client.connect()
        logger.debug(f"Connected to {TOR_CLIENT}") 
                                       
        # Add torrent
        logger.debug(f"Adding {torrent.description.decode('ascii')} with url: {url}")
        client.call("download_torrent_from_url", url)
        print("Torrent sent!")
    except Exception as e:
        print(f"Unable to send to {TOR_CLIENT}. Check the logs for more information.")
        logger.error(f"Error sending to {TOR_CLIENT}. {str(e)}")
        exit()

def transmission(torrent, CLIENT_URL, TOR_CLIENT_USER, TOR_CLIENT_PW, logger):
    TOR_CLIENT = "Transmission"
    print(f"Sending {torrent.description.decode('ascii')} to {TOR_CLIENT}")
    url = fetch_torrent_url(torrent)
    try:
        logger.debug("Connecting to torrent client...")
        # Connection
        if TOR_CLIENT_PW != "":
            logger.debug(f"{TOR_CLIENT} connection info: {CLIENT_URL}, {TOR_CLIENT_USER}")
            client = tClient(address=CLIENT_URL, username=TOR_CLIENT_USER, password=TOR_CLIENT_PW)
        else:
            logger.debug(f"{TOR_CLIENT} connection info: {CLIENT_URL}")
            client = tClient(address=CLIENT_URL)
            logger.debug(f"Connected to {TOR_CLIENT}")    
                                       
        # Add torrent
        logger.debug(f"Adding {torrent.description.decode('ascii')} with url: {url}")
        client.torrent.add(filename=url)
        print("Torrent sent!")
    except Exception as e:
        print(f"Unable to send to {TOR_CLIENT}. Check the logs for more information.")
        logger.error(f"Error sending to {TOR_CLIENT}. {str(e)}")
        exit()

def qbittorrent(torrent, CLIENT_URL, TOR_CLIENT_USER, TOR_CLIENT_PW, logger):
    TOR_CLIENT = "qBittorrent"
    print(f"Sending {torrent.description.decode('ascii')} to {TOR_CLIENT}")
    url = fetch_torrent_url(torrent)
    try:
        logger.debug("Connecting to torrent client...")
        # Connection
        logger.debug(f"{TOR_CLIENT} connection info: {CLIENT_URL}, {TOR_CLIENT_USER}")
        client = qClient(CLIENT_URL)
        client.login(TOR_CLIENT_USER, TOR_CLIENT_PW)
  
                                       
        # Add torrent
        logger.debug(f"Adding {torrent.description.decode('ascii')} with url: {url}")
        client.download_from_link(url)
        print("Torrent sent!")
    except Exception as e:
        print(f"Unable to send to {TOR_CLIENT}. Check the logs for more information.")
        logger.error(f"Error sending to {TOR_CLIENT}. {str(e)}")
        exit()

def local(torrent, download_dir, logger):
    TOR_CLIENT = "local download"
    print(f"Sending {torrent.description.decode('ascii')} to {TOR_CLIENT}")
    url = fetch_torrent_url(torrent)

    try:
        if url.startswith("magnet:?"):
            
            print(f"{torrent.description.decode('ascii')} appears to be a magnet link.")
            logger.info(f"Adding {torrent.description.decode('ascii')} with url: {url}")
            logger.info(f"Using local download method...")   
            convert_to_torrent(url, download_dir)
        else:
            logger.info(f"Using local download method...")   
            r = requests.get(url, allow_redirects=True)
            open(f"{download_dir}{torrent.description}.torrent", 'wb').write(r.content)
            print(f"Torrent downloaded! -> {download_dir}{torrent.description}.torrent")
            logger.info(f"Torrent downloaded! -> {download_dir}{torrent.description}.torrent")
    except Exception as e:
        print(f"Unable to send to {TOR_CLIENT}. Check the logs for more information.")
        logger.error(f"Error sending to {TOR_CLIENT}. {str(e)}")
        exit()


import setuptools
import json
import os
from pathlib import Path
from subprocess import check_call
from setuptools.command.install import install

APP_NAME = "Raincoat"

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        cfg_path = f"{str(Path.home())}/.config/{APP_NAME}.json"
        if not os.path.exists(cfg_path):            
            cfg = open(cfg_path, 'x')
            j = dict(
            jackett_apikey="",
            jackett_url="http://127.0.0.1:9117",
            jackett_indexer="all",
            description_length=100,
            exclude="",
            results_limit=20,
            client_url="",
            display="grid",
            torrent_client="transmission",
            torrent_client_username="",
            torrent_client_password=""
            )
            json.dump(j, cfg, indent=4)
        install.run(self)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raincoat-jackett",
    version="0.8",
    author="Gabisonfire",
    author_email="gabisonfire@github.com",
    description="Raincoat is a tool to search torrents using Jackett and send them to your client.",
    keywords="transmission qbittorrent deluge jackett torrent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    packages=setuptools.find_packages(),
    entry_points = {
        "console_scripts": ['raincoat = raincoat.raincoat:main']
        },
    url="https://github.com/Gabisonfire/raincoat",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Topic :: Communications :: File Sharing",
    ],
    install_requires=["requests==2.23.0", "justlog", "colorama==0.4.3", "tabulate==0.8.7", "transmission-clutch==3.1.0", "deluge-client==1.8.0", "python-qbittorrent==0.4.2"],
    cmdclass={
        'install': PostInstallCommand
    },

)
